@@ -10,19 +10,18 @@ import gc
 
 import numpy as np
 
-import suite2p
-from suite2p.io.binary import BinaryFile
-from lbm_suite2p_python.merging import remake_plane_figures
+from lbm_suite2p_python import default_ops
 from lbm_suite2p_python.postprocessing import (
     ops_to_json,
     load_planar_results,
     load_ops,
-    filter_by_area
 )
 from mbo_utilities.log import get as get_logger
-import mbo_utilities as mbo  # noqa
 
-from lbm_suite2p_python.zplane import save_pc_panels_and_metrics
+from lbm_suite2p_python.zplane import (
+    save_pc_panels_and_metrics,
+    plot_zplane_figures
+)
 
 logger = get_logger("run_lsp")
 
@@ -31,14 +30,12 @@ from lbm_suite2p_python.volume import (
     plot_volume_signal,
     plot_volume_neuron_counts,
     get_volume_stats,
-    plot_execution_time,
 )
-from mbo_utilities.file_io import get_plane_from_filename
+from mbo_utilities.file_io import get_plane_from_filename # derive_tag_from_filename, PIPELINE_TAGS
 
 PIPELINE_TAGS = ("plane", "roi", "z", "plane_", "roi_", "z_")
 
 
-
 def derive_tag_from_filename(path):
     """
     Derive a folder tag from a filename based on “planeN”, “roiN”, or "tagN" patterns.
@@ -90,7 +87,7 @@ def run_volume(
     ops: dict | str | Path = None,
     keep_reg: bool = True,
     keep_raw: bool = False,
-    force_reg: bool = True,
+    force_reg: bool = False,
     force_detect: bool = False,
     dff_window_size: int = 500,
     dff_percentile: int = 20,
@@ -162,6 +159,7 @@ def run_volume(
     - Traces animation over time and neurons
     - Optional rastermap clustering results
     """
+    from mbo_utilities.file_io import get_files, get_plane_from_filename
     start = time.time()
     if save_path is None:
         save_path = Path(input_files[0]).parent
@@ -173,8 +171,8 @@ def run_volume(
     for z, file in enumerate(input_files):
         tag = derive_tag_from_filename(Path(file).name)
         plane_num = get_plane_from_filename(tag, fallback=len(all_ops))
-        subdir = f"plane{plane_num:02d}"
-        plane_save_path = Path(save_path).joinpath(subdir)
+        # subdir = f"plane{tag:02d}"
+        plane_save_path = Path(save_path).joinpath(tag)
         plane_save_path.mkdir(exist_ok=True)
 
         start_file = time.time()
@@ -205,12 +203,16 @@ def run_volume(
 
     if "roi" in Path(input_files[0]).stem.lower():
         print("Detected mROI data, merging ROIs for each z-plane...")
-        from .merging import merge_mrois, remake_plane_figures
+        from .merging import merge_mrois
         merged_savepath = save_path.joinpath("merged_mrois")
         merge_mrois(save_path, merged_savepath)
-        all_ops = sorted(mbo.get_files(merged_savepath, "ops.npy", 2))
+        save_path = merged_savepath
 
-    print(f"Planes found after merge: {len(all_ops)}")
+        all_ops = sorted(get_files(merged_savepath, "ops.npy", 2))
+        print(f"Planes found after merge: {len(all_ops)}")
+    else:
+        all_ops = sorted(get_files(save_path, "ops.npy", 2))
+        print(f"No mROI data detected, planes found: {len(all_ops)}")
 
     try:
         zstats_file = get_volume_stats(all_ops, overwrite=True)
@@ -220,7 +222,7 @@ def run_volume(
             zstats_file, os.path.join(save_path, "mean_volume_signal.png")
         )
         # todo: why is suite2p not saving timings to ops.npy?
-        plot_execution_time(zstats_file, os.path.join(save_path, "execution_time.png"))
+        # plot_execution_time(zstats_file, os.path.join(save_path, "execution_time.png"))
 
         res_z = [
             load_planar_results(ops_path, z_plane=i)
@@ -269,14 +271,14 @@ def run_volume(
 
 def _should_write_bin(ops_path: Path, force: bool = False) -> bool:
     """
-    Decide whether a data_raw.bin should be re-written.
+    Decide whether raw binary export should be performed before registration.
 
     Conditions that trigger re-write:
       - force=True
-      - bin file missing
+      - bin file missing (data.bin or data_chan2.bin)
       - ops.npy missing
       - mismatch between ops metadata (Ly, Lx, nframes) and bin file size
-      - bin file cannot be read or has wrong shape
+      - bin file unreadable or truncated
     """
     if force:
         return True
@@ -284,72 +286,118 @@ def _should_write_bin(ops_path: Path, force: bool = False) -> bool:
     if not ops_path.is_file():
         return True
 
-    bin_path = ops_path.parent / "data.bin"
-    tiff_path = ops_path.parent / "reg_tif"
+    ops = np.load(ops_path, allow_pickle=True).item()
 
-    if not bin_path.is_file() and not tiff_path.is_dir():
-        return True
+    # Check both functional and optional structural binaries
+    bin_candidates = []
+    if "raw_file" in ops:
+        bin_candidates.append(Path(ops["raw_file"]))
+    if "chan2_file" in ops:
+        bin_candidates.append(Path(ops["chan2_file"]))
 
-    try:
-        ops = np.load(ops_path, allow_pickle=True).item()
-        Ly, Lx = ops.get("Ly"), ops.get("Lx")
-        nframes = ops.get("nframes", ops.get("n_frames"))
-
-        if None in (Ly, Lx, nframes):
+    for bin_path in bin_candidates:
+        if not bin_path.is_file():
             return True
+        try:
+            Ly, Lx = ops.get("Ly"), ops.get("Lx")
+            nframes = (
+                ops.get("nframes_chan2")
+                if "chan2" in bin_path.name
+                else ops.get("nframes_chan1", ops.get("nframes"))
+            )
+            if None in (Ly, Lx, nframes):
+                return True
 
-        expected_size = nframes * Ly * Lx * np.dtype(np.int16).itemsize
-        actual_size = bin_path.stat().st_size
+            expected_size = nframes * Ly * Lx * np.dtype(np.int16).itemsize
+            actual_size = bin_path.stat().st_size
+            if actual_size != expected_size:
+                return True
 
-        if actual_size != expected_size:
+            arr = np.memmap(bin_path, dtype=np.int16, mode="r", shape=(nframes, Ly, Lx))
+            _ = arr[0, 0, 0]
+            del arr
+        except Exception as e:
+            print(f"Bin validation failed for {bin_path}: {e}")
             return True
+    return False  # all checks passed
 
-        # Try opening first few frames to verify integrity
-        arr = np.memmap(bin_path, dtype=np.int16, mode="r", shape=(nframes, Ly, Lx))
-        _ = arr[0].sum()  # touch data
-        del arr
+def _should_register(ops_path: str | Path) -> bool:
+    """
+    Determine whether Suite2p registration still needs to be performed.
 
-        return False  # all checks passed
-    except Exception as e:
-        print(f"Bin validation failed: {e}")
-        return True
+    Registration is considered complete if any of the following hold:
+      - A reference image (refImg) exists and is a valid ndarray
+      - meanImg exists (Suite2p always produces it post-registration)
+      - Valid registration offsets (xoff/yoff) are present
+
+    Returns True if registration *should* be run, False otherwise.
+    """
+    ops = load_ops(ops_path)
+
+    has_ref = isinstance(ops.get("refImg"), np.ndarray)
+    has_mean = isinstance(ops.get("meanImg"), np.ndarray)
+    has_offsets = (
+        ("xoff" in ops and np.any(np.isfinite(ops["xoff"]))) or
+        ("yoff" in ops and np.any(np.isfinite(ops["yoff"])))
+    )
+    has_metrics = any(k in ops for k in ("regDX", "regPC", "regPC1", "regDX1"))
+
+    # registration done if any of these are true
+    registration_done = has_ref or has_mean or has_offsets or has_metrics
+    return not registration_done
 
 
 def run_plane_bin(ops) -> bool:
+    from mbo_utilities._binary import BinaryFile
+    from suite2p.run_s2p import pipeline
+    from contextlib import nullcontext
+
     ops = load_ops(ops)
-    if "nframes" in ops and "n_frames" not in ops:
-        ops["n_frames"] = ops["nframes"]
-    if "n_frames" not in ops:
-        raise KeyError("run_plane_bin: missing frame count (nframes or n_frames)")
-    n_frames = ops["n_frames"]
     Ly, Lx = ops["Ly"], ops["Lx"]
 
-    # make sure diam is not nan
-    if ops["diameter"] is not None and np.isnan(ops["diameter"]):
-        ops["diameter"] = 8
-    if ops["diameter"] is None or ops["diameter"] == 0 and ops["anatomical_only"] > 0:
-        ops["diameter"] = 8
-        print("Warning: diameter was not set, defaulting to 8."
-              "Cellpose-SAM currently does not estimate diameter.")
+    # input functional channel (unregistered)
+    raw_file = ops.get("raw_file")
+    nframes_chan1 = ops.get("nframes_chan1") or ops.get("nframes") or ops.get("n_frames")
+    if raw_file is None or nframes_chan1 is None:
+        raise KeyError("Missing raw_file or nframes_chan1")
+
+    # optional structural channel
+    chan2_file = ops.get("chan2_file", "")
+    nframes_chan2 = ops.get("nframes_chan2", 0)
+
+    ops_parent = Path(ops.get("ops_path")).parent
+    ops["save_path"] = ops_parent
+
+    align_structural = ops.get("align_by_chan", 1) == 2
+    ops["align_structural"] = align_structural
+
+    reg_file = ops_parent / "data.bin"
+    ops["reg_file"] = str(reg_file)
+
+    # sanity fix for diameter
+    if "diameter" in ops:
+        if ops["diameter"] is not None and np.isnan(ops["diameter"]):
+            ops["diameter"] = 8
+        if (ops["diameter"] is None or ops["diameter"] == 0) and ops.get("anatomical_only", 0) > 0:
+            ops["diameter"] = 8
+            print("Warning: diameter was not set, defaulting to 8.")
+
     with (
-        suite2p.io.BinaryFile(
-            Ly=Ly, Lx=Lx, filename=ops["reg_file"], n_frames=n_frames
-        ) as f_reg,
-        (
-            suite2p.io.BinaryFile(
-                Ly=Ly, Lx=Lx, filename=ops["raw_file"], n_frames=n_frames
-            )
-            if "raw_file" in ops and ops["raw_file"] is not None
-            else nullcontext()
-        ) as f_raw,
+        BinaryFile(Ly=Ly, Lx=Lx, filename=str(reg_file), n_frames=nframes_chan1) as f_reg,
+        BinaryFile(Ly=Ly, Lx=Lx, filename=raw_file, n_frames=nframes_chan1) as f_raw,
+        (BinaryFile(Ly=Ly, Lx=Lx, filename=chan2_file, n_frames=nframes_chan2)
+        if align_structural else nullcontext()) as f_reg_chan2
     ):
-        ops = suite2p.pipeline(
-            f_reg, f_raw, None, None, ops["do_registration"], ops, stat=None
+        ops = pipeline(
+            f_reg=f_reg,
+            f_raw=f_raw,
+            f_reg_chan2=f_reg_chan2,
+            f_raw_chan2=f_reg_chan2 if align_structural else None,
+            run_registration=ops.get("do_registration", True),
+            ops=ops,
+            stat=None,
         )
-
     np.save(ops["ops_path"], ops)
-    del f_reg, f_raw, ops
-
     return True
 
 
@@ -429,6 +477,8 @@ def run_plane(
     >> output_ops = lsp.run_plane(input_files[0], save_path="D://data//outputs", keep_raw=True, keep_registered=True, force_reg=True, force_detect=True)
     """
     from mbo_utilities.array_types import MboRawArray
+    from mbo_utilities.lazy_array import imread, imwrite
+    from mbo_utilities.metadata import get_metadata
 
     if "debug" in kwargs:
         logger.setLevel(logging.DEBUG)
@@ -457,7 +507,7 @@ def run_plane(
             )
         save_path.mkdir(exist_ok=True)
 
-    ops_default = suite2p.default_ops()
+    ops_default = default_ops()
     ops_user = load_ops(ops) if ops else {}
     ops = {**ops_default, **ops_user, "data_path": str(input_path.resolve())}
 
@@ -466,7 +516,7 @@ def run_plane(
             ops["diameter"]) > 1 and ops["aspect"] == 1.0:
         ops["aspect"] = ops["diameter"][0] / ops["diameter"][1]  # noqa
 
-    file = mbo.imread(input_path)
+    file = imread(input_path)
     if isinstance(file, MboRawArray):
         raise TypeError(
             "Input file appears to be a raw array. Please provide a planar input file."
@@ -474,7 +524,7 @@ def run_plane(
     if hasattr(file, "metadata"):
         metadata = file.metadata  # noqa
     else:
-        metadata = mbo.get_metadata(input_path)
+        metadata = get_metadata(input_path)
 
     if "plane" in ops:
         plane = ops["plane"]
@@ -484,7 +534,7 @@ def run_plane(
         ops["plane"] = plane
     else:
         # get the plane from the filename
-        plane = mbo.get_plane_from_filename(input_path, ops.get("plane", None))
+        plane = get_plane_from_filename(input_path, ops.get("plane", None))
         ops["plane"] = plane
         metadata["plane"] = plane
 
@@ -525,12 +575,10 @@ def run_plane(
             needs_detect = True
 
     ops_file = plane_dir / "ops.npy"
-    reg_data_file = plane_dir / "data.bin"
-    reg_data_file_tiff = plane_dir / "reg_tif"
 
     if _should_write_bin(ops_file, force=kwargs.get("force_save", False)):
         md_combined = {**metadata, **ops}
-        mbo.imwrite(file, plane_dir, ext=".bin", metadata=md_combined, register_z=False)
+        imwrite(file, plane_dir, ext=".bin", metadata=md_combined, register_z=False)
     else:
         print(
             f"Skipping data_raw.bin write, already exists and passes data validation checks."
@@ -542,17 +590,13 @@ def run_plane(
         else {}
     )
 
-    exists = False
-    if reg_data_file.exists():
-        exists = True
-    if reg_data_file_tiff.exists():
-        exists = True
     if force_reg:
         needs_reg = True
     else:
-        # if either reg data file exists, we assume registration is done
-        needs_reg = not exists
-
+        if not ops_file.exists():
+            needs_reg = True
+        else:
+            needs_reg = _should_register(ops_file)
     ops = {
         **ops_default,
         **ops_outpath,
@@ -591,16 +635,18 @@ def run_plane(
         print(f"Skipping {ops_file.name}, processing was not completed.")
         return ops_file
 
-    # cleanup ourselves
-    if not keep_raw:
-        (plane_dir / "data_raw.bin").unlink(missing_ok=True)
-    if not keep_reg:
-        (plane_dir / "data.bin").unlink(missing_ok=True)
+    raw_file = Path(ops.get("raw_file", plane_dir / "data_raw.bin"))
+    reg_file = Path(ops.get("reg_file", plane_dir / "data.bin"))
+
+    if not keep_raw and raw_file.exists():
+        raw_file.unlink(missing_ok=True)
+    if not keep_reg and reg_file.exists():
+        reg_file.unlink(missing_ok=True)
 
     save_pc_panels_and_metrics(ops_file, plane_dir / "pc_metrics")
 
     try:
-        remake_plane_figures(
+        plot_zplane_figures(
             plane_dir,
             dff_percentile=dff_percentile,
             dff_window_size=dff_window_size,
@@ -616,6 +662,8 @@ def run_grid_search(
     grid_search_dict: dict,
     input_file: Path | str,
     save_root: Path | str,
+    force_reg: bool,
+    force_detect: bool,
 ):
     """
     Run a grid search over all combinations of the input suite2p parameters.
@@ -636,6 +684,12 @@ def run_grid_search(
         Root directory where each parameter combination's output will be saved.
         A subdirectory will be created for each run using a short parameter tag.
 
+    force_reg : bool
+        Whether to force suite2p registration.
+
+    force_detect : bool
+        Whether to force suite2p detection.
+
     Notes
     -----
     - Subfolder names for each parameter are abbreviated to 3-character keys and truncated/rounded values.
@@ -688,16 +742,22 @@ def run_grid_search(
             for k, v in combo_dict.items()
         ]
         tag = "_".join(tag_parts)
-
-        print(f"Running grid search in: {save_root.joinpath(tag)}")
-
         save_path = save_root / tag
+        print(f"\nRunning grid search combination: {tag}")
+
+        ops_file = save_path / "ops.npy"
+
+        # Skip runs that are already registered
+        if ops_file.exists() and not force_reg and not _should_register(ops_file):
+            print(f"Skipping {tag}: registration already complete.")
+            continue
+
         run_plane(
             input_path=input_file,
             save_path=save_path,
             ops=ops,
             keep_reg=True,
             keep_raw=True,
-            force_reg=True,
-            force_detect=True,
-        )
+            force_reg=force_reg,
+            force_detect=force_detect,
+        )
\ No newline at end of file
